# Echo client program
import socket
import sys
import struct
from openxm import OpenXMConst
import cmo

class Debug:
    @classmethod
    def out(cls,*args,**kwargs):
        for ai in args:
            print ai,
        print ""
    @classmethod
    def get_line(cls,*args,**kwargs):
        for ai in args:
            print ai,
        r=sys.stdin.readline()
        print r,"=>"
        return r

#####################################################################
class SocketConnection:
    @classmethod
    def try_connecting_as_client(self,host,port):
        s = None
        for res in socket.getaddrinfo(host,port,socket.AF_UNSPEC, socket.SOCK_STREAM):
            (af, socktype, proto, canonname, sa) = res
            try:
                s = socket.socket(af, socktype, proto)
            except socket.error, msg:
                s = None
                continue
            try:
                s.connect(sa)
            except socket.error, msg:
                s.close()
                s = None
                continue
            break
        if s is None:
            Debug.out('could not open socket for %s' % (host+':'+str(port)))
            return s
        else:
            Debug.out('connected to %s' % (host+':'+str(port)))
            return s

    @classmethod
    def try_connecting_as_server(self,host,port):
        s = None
        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM, 0, socket.AI_PASSIVE):
            af, socktype, proto, canonname, sa = res
            try:
                s = socket.socket(af, socktype, proto)
            except socket.error, msg:
                s = None
                continue
            try:
                s.bind(sa)
                s.listen(1)
            except socket.error, msg:
                s.close()
                s = None
                continue
            break
        if s is None:
            Debug.out('could not open socket')
            return None
        conn, addr = s.accept()
        Debug.out('Connected:'+addr)
        return conn


    @classmethod
    def try_connecting_as_server_ipv4(self,host,port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((host, port))
        s.listen(1)
        conn, addr = s.accept()
        Debug.out('Connected:'+addr)
        return conn


class SingleConnectionViaSocket:
    def __init__(self,connected_socket,default_byte_order,cmo_base):
        self.socket=connected_socket
        self.fileobj=self.socket.makefile()
        self.set_byte_order(default_byte_order)
        self.serial=0
        self.cmo_base=cmo_base
        self.set_available_sm_command(None)
        self.set_available_data(self.cmo_base.get_implemented_cmo())


    def close(self):
        if self.fileobj:
            self.fileobj.close()
            self.fileobj=None
        if self.socket:
            self.socket.close()
            self.socket=None

    def negotiate_byte_order(self,is_oxclient):
        if is_oxclient:
            data=self.read_byte()
            self.set_byte_order(data)
            self.send(struct.pack('B',self.byte_order))
            return self.byte_order
        else:
            self.send(struct.pack('B',self.byte_order))
            data=self.read_byte()
            self.set_byte_order(data)
            return self.byte_order
    def set_byte_order(self,byteorder):
        if byteorder == OpenXMConst.OX_BYTE_NETWORK_BYTE_ORDER:
            self.set_byte_order_char('!')
            self.byte_order=OpenXMConst.OX_BYTE_NETWORK_BYTE_ORDER
        elif byteorder == OpenXMConst.OX_BYTE_BIG_ENDIAN:
            self.set_byte_order_char('>')
            self.byte_order=OpenXMConst.OX_BYTE_BIG_ENDIAN
        elif byteorder == OpenXMConst.OX_BYTE_LITTLE_ENDIAN:
            self.set_byte_order_char('<')
            self.byte_order=OpenXMConst.OX_BYTE_LITTLE_ENDIAN
        else:
            self.set_byte_order_char('!')
            self.byte_order=OpenXMConst.OX_BYTE_NETWORK_BYTE_ORDER
        return self.byte_order
    def set_byte_order_char(self,byteorderchar):
        self.byte_order_char=byteorderchar
        self.format_int32=self.byte_order_char+'i'

    def set_available_sm_command(self,list_of_sm_command_tags):
        if list_of_sm_command_tags==None:
            self.available_sm_commands=None
        else:
            self.available_sm_commands=list_of_sm_command_tags[:]
    def set_available_data(self,list_of_data_tags):
        if list_of_data_tags==None:
            self.available_data=None
        else:
            self.available_data={}
            for (k,list_of_tags) in list_of_data_tags:
                self.available_data[k]=list_of_tags[:]
    def is_available_sm_command(self,sm_tag):
        if self.available_sm_commands:
            if sm_tag in self.available_sm_commands:
                return True
            else:
                return False
        else:
            return True
    def is_available_data(self,ox_tag,cmo_tag):
        if self.available_data:
            if ox_tag in self.available_data:
                if cmo_tag in self.available_data[ox_tag]:
                    return True
            return False
        else:
            return True


    def send(self,bytes):
        self.socket.send(bytes)
    def recv(self,bufsize):
        return self.socket.recv(bufsize)
    def read(self,size):
        return self.fileobj.read(size)


    def read_byte(self):
        return self.fileobj.read(1)
    def read_int32(self):
        return struct.unpack(self.format_int32,self.fileobj.read(4))[0]
    def read_cmo(self):
        return self.cmo_base.new_from_stream(self,self.byte_order_char)
    def read_ox_tag(self):
        tag=self.read_int32()
        serial=self.read_int32()
        return (tag,serial)


    def read_ox_expression(self):
        (tag,serial)=self.read_ox_tag()
        if tag==OpenXMConst.OX_COMMAND:
            body=self.read_int32()
        elif tag==OpenXMConst.OX_DATA:
            body=self.read_cmo()
        else:
            body=None
        return (tag,serial,body)

    def send_byte(self,byte):
        self.send(struct.pack('B',byte))
    def send_int32(self,n):
        self.send(struct.pack(self.format_int32,n))
    def send_cmo(self,cmobj):
        self.send(cmobj.to_byte(self.byte_order_char))

    def send_ox_tag(self,tag):
        self.serial=self.serial+1
        self.send_int32(tag)
        self.send_int32(self.serial)

    def send_ox_command(self,function_id):
        if self.is_available_sm_command(function_id):
            self.send_ox_tag(OpenXMConst.OX_COMMAND)
            self.send_int32(function_id)
            Debug.out(">> sent the command:",function_id)
        else:
            Debug.out(">> cannot send the command:",function_id)
    def send_ox_cmodata(self,oxtag,cmobj):
        if self.is_available_data(oxtag,cmobj.get_cmo_tag()):
            self.send_ox_tag(oxtag)
            self.send_cmo(cmobj)
            Debug.out(">> sent the data:",oxtag,cmobj.to_python_object())
        else:
            Debug.out(">> cannot send the data:",oxtag,cmobj.to_python_object())
    def send_ox_data(self,cmobj):
        self.send_ox_cmodata(OpenXMConst.OX_DATA,cmobj)


    def send_ox_command_and_read_output(self,function_id):
        self.send_ox_command(function_id)
        return self.read_ox_expression()        


    def send_sm_popString_and_read(self):
        return self.send_ox_command_and_read_output(OpenXMConst.SM_popString)
    def send_sm_popCMO_and_read(self):
        return self.send_ox_command_and_read_output(OpenXMConst.SM_popCMO)
    def send_sm_popSerializedLocalObject_and_read(self):
        return self.send_ox_command_and_read_output(OpenXMConst.SM_popSerializedLocalObject)


            
class DuplicateConnectionForOXSession:
    def __init__(self,host,port_control,port_data,cmo_base_control,cmo_base_data):
        self.host=host
        self.port_control=port_control
        self.port_data=port_data
        self.control=None
        self.data=None
        self.cmo_base_control=cmo_base_control
        self.cmo_base_data=cmo_base_data

    def connect(self):
        socket_control=SocketConnection.try_connecting_as_client(self.host,self.port_control)
        if socket_control:
            socket_data=SocketConnection.try_connecting_as_client(self.host,self.port_data)
            if socket_data:
                self.control=SingleConnectionViaSocket(socket_control,0,self.cmo_base_control)
                self.data=SingleConnectionViaSocket(socket_data,0,self.cmo_base_data)
                return True
            else:
                socket_control.close()
        return False

    def close(self):
        if self.control:
            self.control.close()
            self.control=None
        if self.data:
            self.data.close()
            self.data=None

    def negotiate_byte_order(self,is_oxclient,control_before_data=True):
        if control_before_data:
            byte_order_control=self.control.negotiate_byte_order(is_oxclient)
            byte_order_data=self.data.negotiate_byte_order(is_oxclient)
        else:
            byte_order_data=self.data.negotiate_byte_order(is_oxclient)
            byte_order_control=self.control.negotiate_byte_order(is_oxclient)
        return (byte_order_control,byte_order_data)
    def set_available_sm_command(self,list_of_sm_command_tags_control,list_of_sm_command_tags_data):
        self.control.set_available_sm_command(list_of_sm_command_tags_control)
        self.data.set_available_sm_command(list_of_sm_command_tags_data)
    def set_available_data(self,list_of_data_tags_control,list_of_data_tags_data):
        self.control.set_available_data(list_of_data_tags_control)
        self.data.set_available_data(list_of_data_tags_data)


class OpenXMSimpleClient:
    def __init__(self,host,port_ctrl,port_data):
        self.connection=DuplicateConnectionForOXSession(host,port_ctrl,port_data,cmo.cmo,cmo.cmo) 

    def start_session(self):
        session_established=self.connection.connect()
        if session_established:
            (byte_order_control,byte_order_data)=self.connection.negotiate_byte_order(True)
            Debug.out("Byte order of control:",byte_order_control)
            Debug.out("Byte order of data:",byte_order_data)
        return session_established
            
    def end_session(self):
        self.connection.data.send_ox_command(OpenXMConst.SM_shutdown)
        self.connection.data.read(-1)
        self.connection.close()

    def get_mathcap(self):
        self.connection.data.send_ox_command(OpenXMConst.SM_mathcap)
        (tag,ser,mathcap)=self.connection.data.send_sm_popCMO_and_read()
#        sysinfo=mathcap.body.get_sequence()[0]
        (s,sm,da)=mathcap.to_python_object()
        return (s,sm,da)

    def negotiate_mathcap(self):
        (systeminfo,availablesm,availableoxdata)=self.get_mathcap()
        sm_data=[ sm_tag for sm_tag in availablesm if self.connection.data.is_available_sm_command(sm_tag)] 
        sm_control=[ sm_tag for sm_tag in availablesm if self.connection.control.is_available_sm_command(sm_tag)]
        d=[(k, [cmo_tag for cmo_tag in tags if self.connection.data.is_available_data(k,cmo_tag)]) for (k,tags) in availableoxdata]
        data_data=[(k,l) for (k,l) in d if l!=[] ]
        d=[(k, [cmo_tag for cmo_tag in tags if self.connection.control.is_available_data(k,cmo_tag)]) for (k,tags) in availableoxdata]
        data_control=[(k,l) for (k,l) in d if l!=[] ]
        self.connection.set_available_sm_command(sm_control,sm_data)
        self.connection.set_available_data(data_control,data_data)
        sm=cmo.CList([cmo.CInt32(i) for i in sm_data])
        da=cmo.CList([cmo.CList([cmo.CInt32(k),cmo.CList([cmo.CInt32(li) for li in l])]) for (k,l) in data_data])
        mc=cmo.CMathcap(cmo.CList([cmo.CList([]),sm,da]))
        self.connection.data.send_ox_data(mc)
        self.connection.data.send_ox_command(OpenXMConst.SM_setMathCap)
        return (systeminfo,sm_data,data_data)


######################################################################


def test_main():
    oxmsc=OpenXMSimpleClient("localhost",1200,1300)
    session_established=oxmsc.start_session()
    if not session_established:
        Debug.out('Connection error.')
        return 
    r=oxmsc.negotiate_mathcap()
    Debug.out('Mathcap:',r)
#    Debug.get_line('Please type return key:')
#    r=oxmsc.get_mathcap()
#    Debug.out('Current MathCap:',r)
    Debug.out("")
    r=Debug.get_line('Please input string to excute (q for quit):')
    while r != "q" and r != "q\n": 
        oxmsc.connection.data.send_ox_data(cmo.CString(r))
        oxmsc.connection.data.send_ox_command(OpenXMConst.SM_executeStringByLocalParser)
        r=Debug.get_line('Pop? s(pop as string)/x(pop as cmo and print as xml)/y(both)/other(No):')
        if r=="y" or r=="y\n":
            cmobj=oxmsc.connection.data.send_sm_popCMO_and_read()[-1]
            r=cmobj.to_xml()
            Debug.out(r)
            oxmsc.connection.data.send_ox_data(cmobj)
            r=oxmsc.connection.data.send_sm_popString_and_read()
            Debug.out(r[-1].get_sequence())
        elif r=="s" or r=="s\n":
            r=oxmsc.connection.data.send_sm_popString_and_read()
            Debug.out(r[-1].get_sequence())
        elif r=="x" or r=="x\n":
            cmobj=oxmsc.connection.data.send_sm_popCMO_and_read()[-1]
            r=cmobj.to_xml()
            Debug.out(r)
        Debug.out("")
        r=Debug.get_line('Please input string to excute (q for quit):')
    Debug.get_line('Please type return key to close this session.')
    oxmsc.end_session()



if __name__ == "__main__":
    test_main()



